//: sequence.hpp : Sequenz-Klasse - R.Richter 2002-10-14
/////////////////////////////////////////////////////////////////////

#ifndef SEQUENCE_HPP
#define SEQUENCE_HPP
/// === Aufgabe
/// Idee aus:
/// [Ruminations] Andrew Koenig, Barbara Moo: Ruminations on C++.
/// Addison-Wesley, Reading, Mass. (1997) 175-190.

/// Gesucht ist eine Klassenschablone fr Datenfolgen.

/// Eine Folge (Sequenz) ist ein sehr einfaches, 
/// aber leistungsfhiges Konzept fr die Datenorganisation. 
/// Schlielich sind ganze Programmiersprachen darauf aufgebaut 
/// (LISP = List Processing).
/// Dieser Quelltext macht Sequenzen in C++ nutzbar.

/// Anforderungen an Sequenz:
// Sequence();             	leere Sequenz bauen (Standardkonstruktor)
// Sequence(head, tailseq);	Sequenz aus neuen Kopfwert und Schwanzfolge
// bool    empty() const;  	wahr, wenn keine weiteren Werte in Sequenz
// T        head() const;  	Kopfwert;  Sequenz darf nicht leer sein
// Sequence tail() const;  	Restfolge; Sequenz darf nicht leer sein
///
/// === Lsung
/// Falls etwas schiefgeht, teilen wir das mit einer Ausnahmeklasse mit:

class SequenceError
{
public:
  SequenceError(const char* message) : msg(message) {}
  virtual const char* what() const { return msg; }
private:
  const char* msg;
};

/// Die Sequenzklasse ist als Schablone zu definieren.
/// Eine Folge enthlt stets Elemente des gleichen Typs, 
/// von Folge zu Folge kann der Typ jedoch anders sein.

template <class T>
class Sequence
{
private:
  struct Item; 
  Item* item;

/// Eintrge (Items) werden in einer privaten Struktur gehalten. 
/// Fr den Anfang geben wir uns mit einer Forward-Deklaration zufrieden.
/// Nur der erste Eintrag der Folge wird ber einen Zeiger verwaltet.
/// Die Sequenz-Inhalte bilden eine einfach verkettete Liste.
/// Der Eintrag enthlt einen Wert und einen Zeiger auf nchsten Eintrag.

  struct Item
  {
    const T value;
    Item* next;
    int use;       // reference count

/// Die Werte der Folge sind nicht nderbar (const). 
/// Wir knnen das Kopieren von ganzen oder von Teilfolgen vermeiden
/// durch die gemeinsame Nutzung der darunterliegenden Sequenz.
/// Das erfordert Referenzzhler.

/// Der einzige trickreiche Teil der Implementierung betrifft 
/// das Bauen einer neuen Sequenz aus einem neuen Wert x 
/// und einer existierenden Sequenz. Danach mssen beide Sequenzen 
/// mit einer gemeinsamen Restfolge auskommen:

    Item(const T& x, Item* successor)
      : use(1), value(x), next(successor)
    {
      if(successor) successor->use++;
    }
  };

/// Wichtig ist, die Referenzzhlerei richtig zu verstehen. 
/// Der Nachfolger wird dann einmal mehr benutzt,
/// der erste ist neu, also nur einmal gezhlt.
/// Beachte: Der Zhler beim ersten kann niedriger sein als beim zweiten.
/// Das wird beim weiteren Bau der Sequenzklasse wichtig sein.

public:

/// Der Standardkonstruktor definiert eine leere Folge.
/// Wir setzen den item Zeiger auf Null:

  Sequence() : item(0) {}

/// Eine Sequenz aus Kopfwert und Restfolge zusammenzusetzen
/// ist kaum schwieriger:

  Sequence(const T& headvalue, const Sequence& tailseq) 
    : item(new Item(headvalue, tailseq.item)) 
  {}

/// Da wir Zeiger verwalten, mssen wir den Kopierkonstruktor angeben.
/// Wir mssen nur daran denken, den Referenzzhler des Anfangs zu erhhen:

  Sequence(const Sequence& original)
    : item(original.item)
  {
    if(item) item->use++;
  }

/// Destruktor und Zuweisungsoperator mssen mglicherweise 
/// eine ganze Folge freigeben.
/// Wir kommen spter darauf zurck, wenn die einfachen Dinge fertig sind.
/// Wann die Sequenz leer ist, ist offensichtlich:

  bool empty() const { return item==0; }

  T head() const
  {
    if(!item) throw SequenceError("no head in empty sequence");
    return item->value;  
  }  

  Sequence tail() const
  {
    if(!item) throw SequenceError("no tail in empty sequence");
    return Sequence(item->next);  
  }

/// Kopf und Rest der Folge lassen sich nur bilden, 
/// wenn die Folge mindestens einen Wert enthlt.
/// Falls nicht, mssen wir eine Ausnahme werfen.
/// Der Rest der Liste muss aus item->next gebildet werden, 
/// aber dies ist ein Item, wir brauchen jedoch eine neue Sequence.
/// Dafr brauchen wir noch einen (privaten) Konstruktor:

private:
  Sequence(Item* entry)
    : item(entry)
  {
    if(item) item->use++;
  }

/// Wir wenden uns wieder der Zuweisung und dem Destruktor zu.
/// Den komplizierten Teil des Abbauens einer Folge 
/// verstecken wir in einer (privaten) Hilfsmethode,
/// deren Implementation im Anschluss an die Klasse erfolgt.
/// Methoden, die Schleifen enthalten, 
/// lassen sich auf vielen Compilern nicht inline implementieren.

  static void destroy(Item* item);

public:
  ~Sequence() { destroy(item); }  

  Sequence& operator=(const Sequence& original) 
  {
    if(original.item) original.item->use++;
    destroy(item);
    item = original.item;
    return *this;
  }

/// Der Zuweisungsoperator muss sich gegen Selbstzuweisung schtzen.
/// Wir erhhen deshalb den Zhler der rechten Seite, 
/// bevor die linke Seite abgebaut wird.
/// Alternative (Standard-)Implementation htte das Schema
/*
    if(this!=&original)      // falls nicht dieselbe Sequenz
    {
      destroy(item);         // rume linke Seite auf
      item = original.item;  // klinke rechts ein
      if(item) item->use++;  // erhhe den Zhler
    }
    return *this;
*/

/// Fr C- und C++-Programmierer ist eine zeigerartige Schreibweise 
/// vertraut, auch wenn es sich bei den Sequenzen genaugenommen 
/// nicht um Zeiger handelt:

/*
  for(Seq s2=s; s2; ++s2)
  {
    cout << *s2 << ' ';  // 1 2 3 4 5
  }
*/

/// Die Implementierung ist einfach:

// pointer-like metapher - overloading operators (member)
   
  operator bool() const { return item!=0; }
  T operator*()   const { return head(); }
  Sequence& operator++()   { return *this=tail(); } 
  Sequence operator++(int) { Sequence tmp=*this; ++(*this); return tmp; }

/// Der Postfix-Operator (erkennbar am nicht benutzten int-Parameter)
/// muss das gewohnte Verhalten der Postfix-Operatoren nachbilden
/// und die alte Sequenz zurckgeben.

/// Lsst sich einer Sequenz eine durch Komma getrennte Liste von Werten 
/// zuweisen?
/*
  Sequence s;
  s = 1, 2, 3, 4, 5;
*/
/// Aus der Numerik-Bibliothek Blitz++ fr mehrdimensionale Arrays
/// ist zu erkennen, dass und wie es geht. 

/// [Blitz++] Todd Veldhuizen: Techniques for Scientific C++. PDF-Datei.

/// Dazu wird eine neue Hilfsklasse, ein Initialisierer bentigt.
/// Der Initialisierer kann davon ausgehen, 
/// immer auf dem letzten Knoten einer neuen Sequenz zu arbeiten.
/// Mit dem Initialisierer wird an der Liste entlanggehangelt.
/// Dazu wird (sehr ungewhnlich!) der Kommaoperator berladen.

  // list initializer
  class Init
  {
  public:
    Init(Item* lastnode) : item(lastnode) {}
    Init operator,(const T& x) 
    {
      item->next = new Item(x, 0);
      return Init(item->next);
    }
  private:  
    Item* item; 
  };

/// Wird versucht, einer Sequenz einen einzelnen Wert zuzuweisen,
/// ist das als Beginn einer Initialisiererliste zu werten.

  Init operator=(const T& x)
  {
    destroy(item);
    item = new Item(x, 0);
    return Init(item);
  }  
};

/// Damit ist die Klassendefinition abgeschlossen.
/// Zugegeben, das berladen des Kommaoperators 
/// ist eine sehr gefhrliche Sache.
/// Unabsichtliche Zuweisungen lschen die enthaltene Folge.
/// Darber hinaus ist es nicht mglich, 
/// die Liste beim Konstruktor anzugeben,
/// da ein Konstruktor keinen Wert zurckgegeben kann.

/// Es bleibt, die aufgeschobene @code{destroy()}-Methode zu implementieren.
/// Die Aufrumarbeiten mssen sorgfltig erledigt werden.
/// Die Folge darf nur berhaupt und nur soweit abgerumt werden,
/// wie dieses Objekt der einzige Besitzer ist.

template <class T>
void Sequence<T>::destroy(Sequence::Item* item)
{
  while(item && --item->use==0)
  {
    Item* next = item->next;
    delete item;
    item = next;
  }
}

/// Globale Funktionen knnen die Sequenzklasse so umhllen, 
/// dass die Sequenz Lisp-artig funktional genutzt werden kann:

// Lisp like global functions:
template <class T>
inline
bool empty(const Sequence<T>& s) // Lisp: null?
{
  return s.empty();
}

template <class T>
inline
T head(const Sequence<T>& s) // Lisp: car
{
  return s.head();
}

template <class T>
inline
Sequence<T> tail(const Sequence<T>& s) // Lisp: cdr
{
  return s.tail();
}

template <class T>
inline
Sequence<T> 
concat(const T& x, const Sequence<T>& s) // Lisp: cons
{
  return Sequence<T>(x, s);
}

template <class T>                       // Lisp:
Sequence<T> 
concat(Sequence<T> s, const T& x)        // (define append  
{                                        //   (lambda s x)
  if(empty(s))                           //   (if (null? s)
  {                                      //  
    return concat(x, Sequence<T>());     //       (cons x '())
  }                                      //       
  return concat(s.head(),                //       (cons (car s) 
                concat(s.tail(), x)      //             (append (cdr s) x))
               );                        //   )
}                                        // )  

#endif // SEQUENCE_HPP
